/* OpenCA libpki package
* (c) 2000-2007 by Massimiliano Pala and OpenCA Group
* All Rights Reserved
*
* ===================================================================
* Released under OpenCA LICENSE
*/

#include <libpki/pki.h>

#define BUFF_MAX_SIZE	2048

HTTP_RV *parse_http_headers_mem ( PKI_MEM *pki_mem ) {

        int head_lines = 0;
        char linebuf[1024];

	HTTP_RV *ret = NULL;

	BIO *in = NULL;

	if( !pki_mem || !pki_mem->data || pki_mem->size < 1 ) {
		return (NULL);
	}

	if( (in = BIO_new_mem_buf( pki_mem->data, (int) pki_mem->size )) 
								== NULL ) {
		return ( NULL );
	}

	if((ret = PKI_Malloc ( sizeof( HTTP_RV ))) == NULL ) {
		return ( NULL );
	}

	ret->code = 0;
	ret->size = 0;
	ret->location = NULL;

        head_lines = 0;
        do {
                int read_code = 0;
		char *eol = NULL;

                memset(linebuf, '\x0', sizeof (linebuf));
                if((read_code = BIO_gets(in, linebuf, sizeof(linebuf))) < 0) {
                        PKI_log_debug( "HTTP, Error (%d) retrieving "
					"data (head_lines=%d)", 
                                                read_code, head_lines );
			if( ret ) PKI_Free ( ret );
                        return (NULL);
                }

		ret->size += (int) strlen(linebuf);

		if( strlen(linebuf) < 1 ) {
			break;
		}

		if( (eol = strchr ( linebuf, '\r' )) ||
				(eol = strchr( linebuf, '\n' ))) {
			*eol = '\x0';
		}

                /* Let's check if the first line has good code */
		if( (head_lines == 0 ) && 
			( (ret->code = atoi(&linebuf[9])) != 200 ) ) {
			/* Not really found - let's return the
 			   code, just negative to distinguish from
			   the normal case */
			PKI_log_debug("HTTP returned code (%d)", ret->code );
		}

		if( strncmp_nocase("Location:", linebuf, strlen("Location:"))
									== 0 ){
			char *pnt = NULL;

			if((pnt = strchr( linebuf, ' ' )) == NULL ) {
				pnt = strchr ( linebuf, ':' ) + 1;
			}
			while ( (pnt != NULL ) && ( *pnt == ' ' ) ) {
				pnt++;
			};

			if( pnt ) {
				size_t len = 0;

				len = strlen( pnt );

				ret->location = PKI_Malloc ( len + 1 );
				memcpy( ret->location, pnt, len );
				ret->location[len] = '\x0';

				// PKI_log_debug( "DEBUG::LOCATION::%s", 
				// 				ret->location );
			}
		}

		head_lines++;
        }
        while( strlen(linebuf) > 0 );

	// PKI_log_debug(">>>>>> ORIGINAL DATA <<<<<\n%s\n>>>>>>>>>><<<<<<<<\n",
	// 		PKI_MEM_get_data( pki_mem ));
	BIO_free( in );

	return ( ret );
}

PKI_MEM *http_get_data_sock ( int fd, int timeout, ssize_t max_size ) {

        PKI_MEM *buf = NULL;

	char tmp_buff[BUFF_MAX_SIZE];

        ssize_t newsize  = 0;

        if( fd < 1 ) {
		PKI_log_err("Attempted to retrieve data from sock %d", fd );
		return NULL;
	}

        /* get subject name from bio using recommended OpenSSL template */
        if((buf = PKI_MEM_new_null()) == NULL ) {
		PKI_log_err( "Memory Failure" );
		return ( NULL );
	}

	while( (newsize = PKI_NET_read( fd, tmp_buff, sizeof(tmp_buff) , timeout )) != 0 ) {

		if( newsize < 0 ) {
			PKI_log_err("Network Error: %s", strerror(errno));
			break;
		}

		if( (max_size > 0) && 
				((ssize_t)(buf->size) + newsize > max_size) ) {
			newsize = max_size - (ssize_t) (buf->size);
			PKI_MEM_add( buf, tmp_buff, (size_t) newsize);
			break;
		};

		PKI_MEM_add ( buf, tmp_buff, (size_t) newsize );

	};

	if( buf->size <= 0 ) {
		PKI_log_debug("WARNING::No HTTP data retrieved.");

		if( buf ) PKI_MEM_free ( buf );
		buf = NULL;
	}

	return ( buf );
}

/*! \brief Sends (POST/GET) data to a url and (if a pointer to a mem stack
 *         is provided) returns the received response. PKI_ERR is returned
 *         in case of error, otherwise PKI_OK is returned.
 */

int URL_get_data_http_url ( URL *url, char *data, ssize_t data_size,
			int method, char *content_type, PKI_MEM_STACK **sk,
				int timeout, ssize_t max_size ) {

	int socket          = -1;

	ssize_t len          = 0;

	PKI_MEM *obj 	    	= NULL;
	PKI_MEM_STACK *data_sk  = NULL;

	char *my_cont_type	= "application/unknown";

	HTTP_RV *http_rv    = NULL;
	int	      rv    = -1;
	int 	     ret    = PKI_OK;

	size_t max_len = 0;

	char *tmp  = NULL;
    
	char *head_get =    "GET %s HTTP/1.0\r\n"
			"Host: %s\r\n"
			"User-Agent: LibPKI\r\n"
		        "Connection: close\r\n\r\n"
			"\r\n";
	char *head_post =    "POST %s HTTP/1.0\r\n"
			"Host: %s\r\n"
			"User-Agent: LibPKI\r\n"
		        "Connection: close\r\n"
			"Content-type: %s\r\n"
			"Content-Length: %d\r\n"
			"\r\n";

	char *head = NULL;

	if ( timeout < 0 ) timeout = 0;

	PKI_log_debug ("HTTP::Method %d::Connecting to => %s", method, url->addr );

	if((socket = PKI_NET_open ( url, 0 )) < 0 ) {
		/* Can not connect to the server */
		PKI_log_err ("Failed to connect to %s:%d", url->addr, url->port);
		return PKI_ERR;
	}

	if ( method == PKI_HTTP_METHOD_GET ) {
		head = head_get;
		max_len = strlen(head) + strlen( url->path ) + strlen( url->addr ) + 1;
		tmp = PKI_Malloc ( max_len );
		len = (ssize_t) snprintf(tmp, max_len, head, url->path, url->addr );
	} else if ( method == PKI_HTTP_METHOD_POST ) {
		head = head_post;
		if ( content_type ) my_cont_type = content_type;
		max_len = strlen(head) + strlen( url->path ) + 
				strlen( url->addr ) + strlen(my_cont_type) + 101;
		tmp = PKI_Malloc ( max_len );
		len = (ssize_t) snprintf(tmp, max_len, head, url->path, url->addr, 
					my_cont_type, data_size );
	} else {
		PKI_log_err ( "Method (%d) not supported!", method );
		PKI_NET_close ( socket );
		return PKI_ERR;
	}

	// PKI_log_debug ( "HEADERS::\n%s\n", tmp );

	if((rv = PKI_NET_write ( socket, tmp, (size_t) len )) < 0 ) {
		/* Error in SSL_write */
		PKI_log_err("Can not write HTTP header to socket");
		goto err;
	}

	if( method == PKI_HTTP_METHOD_POST ) {
		/* Send the data together with the header */
		if((rv = PKI_NET_write ( socket, data, data_size )) < 0 ) {
			PKI_log_err ("Can not write POST data to socket.");
			goto err;
		}
	}
		

	if((obj = http_get_data_sock( socket, timeout, max_size )) == NULL ) {
		PKI_log_debug( "No data returned from SSL connection");
		goto err;
	};

	// fprintf(stderr, "[%d] RECEIVED_DATA => %s\n", obj->size, obj->data );

	if( (http_rv = parse_http_headers_mem ( obj )) == NULL ) {
		PKI_log_err("Unknown HTTP Error");
		goto err;
	}

	if ( http_rv->code >= 400 ) {
		goto err;
	} else if ( http_rv->code >= 300 ) {
		/* Redirection - let's try that */
		char new_url[2048];

		if( !http_rv->location ) {
			PKI_log_debug ( "HTTP Redirection but no location "
						"provided!");
			goto err;
		}

		if( strstr(http_rv->location, "://") != NULL ) {

			
			if( strncmp_nocase( http_rv->location,
				    url->url_s, (int) strlen(http_rv->location)) == 0 ) {
				PKI_log_debug( "HTTP cyclic redirection!");
				goto err;
			}

			ret = URL_get_data_http_url ( url, data, data_size,
                        		method, content_type, sk,
                                		timeout, max_size );
			goto end;

		} else {
			const char *prot_s = NULL;
			URL *my_new_url = NULL;

			prot_s = URL_proto_to_string ( url->proto );
			if( !prot_s ) goto err;

			snprintf(new_url, sizeof(new_url),"%s://%s%s",
				prot_s, url->addr, http_rv->location );

			PKI_log_debug("OLD URL [%s] --> NEW URL [%s]",
				url->url_s, new_url );

			if( strncmp_nocase( new_url,
					url->url_s, (int) strlen(new_url)) ==0){
				PKI_log_debug( "HTTP cyclic redirection!");
				goto err;
			}

			if ( http_rv ) {
				if ( http_rv->location ) PKI_Free ( http_rv->location );
				PKI_Free ( http_rv );
			}

			if( obj ) PKI_MEM_free ( obj );

			my_new_url = URL_new ( new_url );

			ret = URL_get_data_http_url ( my_new_url, data, data_size,
					method, content_type, sk, timeout,
						max_size );

			goto end;
		}
	} else if ( http_rv->code != 200 ) {
		PKI_log_debug( "HTTP Return code not manageable (%d)",
							http_rv->code );
		goto err;
	}

	/* If a Pointer was provided, we want the data back :D */
	if ( sk ) {
		PKI_MEM *mm = NULL;

		if ( ! *sk ) {
			if((*sk = PKI_STACK_MEM_new()) == NULL ) {
				PKI_log_debug ("Memory Failure");
				goto err;
			}
		}

		if((mm = PKI_MEM_new_null()) == NULL ) {
			PKI_log_debug ("Memory Failure");
			goto err;
		}

		/* Add the data to the MEM */
		PKI_MEM_add( mm, (char *) &(obj->data[http_rv->size]), 
			(size_t) (obj->size) - (size_t) (http_rv->size) );

		/* Add the returned value to the stack */
		PKI_STACK_MEM_push ( *sk, mm );
	}

end:
	if ( http_rv ) {
		if ( http_rv->location ) PKI_Free ( http_rv->location );
		PKI_Free ( http_rv );
	}

	return ret;

err:
	if ( tmp ) PKI_Free ( tmp );
	if ( obj ) PKI_MEM_free ( obj );

	if ( http_rv ) {
		if ( http_rv->location ) PKI_Free ( http_rv->location );
		PKI_Free ( http_rv );
	}

	return PKI_ERR;
}

/*! \brief Returns the data from an HTTP source by using the GET command */

PKI_MEM_STACK *URL_get_data_http ( char *url_s, int timeout, ssize_t max_size ) {
	URL *url = NULL;
	PKI_MEM_STACK *ret = NULL;

	if( !url_s ) return (NULL);

	if((url = URL_new( url_s )) == NULL ) {
		return NULL;
	}

	if( url->proto != URI_PROTO_HTTP ) {
		if( url ) URL_free ( url );
		return( NULL );
	}

	if (( URL_get_data_http_url ( url, NULL, -1, PKI_HTTP_METHOD_GET,
			NULL, &ret, timeout, max_size )) == PKI_ERR ) {
		return NULL;
	}

	return ret;
}

/*! \brief Use POST method to transfer data via HTTP. If a pointer to a
 *         PKI_MEM_STACK (eg., &stack) is provided, the returned data isi
 *         added to it
 */

int URL_post_data_http ( char *url_s, char *data, ssize_t size, 
				char *content_type, PKI_MEM_STACK **ret_sk,
					int timeout, ssize_t max_size ) {

	URL *url = NULL;
	int ret = PKI_OK;

	if( !url_s || !data || !content_type ) {
		/* ERROR: All data are strictly required! */
		return(PKI_ERR);
	}

	if((url = URL_new(url_s)) == NULL ) {
		/* Error in creating the URL structure */
		return(PKI_ERR);
	}

	ret = URL_post_data_http_url ( url, data, size, content_type,
				ret_sk, timeout, max_size );

	if ( url ) URL_free (url);

	return ret;
}

/*! \brief Posts data to an HTTP URL and returns the response data */

int URL_post_data_http_url ( URL *url, char *data, ssize_t size, 
				char *content_type, PKI_MEM_STACK **ret_sk, 
					int timeout, ssize_t max_size ) {

	int ret = PKI_OK;

	if ( !url || !data || size <= 0 ) return PKI_ERR;

	return URL_get_data_http_url ( url, data, size,
                        PKI_HTTP_METHOD_POST, content_type, ret_sk,
                                timeout, max_size );

}
