/*
 * LIBPKI - Easy PKI Library
 * by Massimiliano Pala (madwolf@openca.org)
 * OpenCA project 2007
 *
 * Copyright (c) 2007 The OpenCA Project.  All rights reserved.
 *
 * ====================================================================
 *
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).  This product includes software written by Tim
 * Hudson (tjh@cryptsoft.com).
 *
 */

#include <libpki/pki.h>

#ifdef HAVE_LDAP


LDAP *URL_LDAP_connect( URL *url ) {

	LDAP	*ld = NULL;
	int	protocol = -1;
	int	ret = 0;

	
        (void) signal( SIGPIPE, SIG_IGN );

	if( (!url) || (!url->addr) || (!url->path)) {
		return NULL;
	}

	if ((ld = ldap_init( url->addr, url->port )) == NULL ) {
		return ( NULL );
	}

	/* Setting LDAP protocol V3 */
	protocol = LDAP_VERSION3;
	if( ldap_set_option( ld, LDAP_OPT_PROTOCOL_VERSION, &protocol )
						!= LDAP_SUCCESS ) {
		ldap_unbind(ld);
		return(NULL);
	}

	if ( (ret = ldap_bind_s( ld, url->usr, 
				url->pwd, LDAP_AUTH_SIMPLE )) 
		!= LDAP_SUCCESS ) {
		PKI_log_err( "ldap_bind_s failure (%s:%d) [%d]",
			     url->addr, url->port, ret );
		return NULL;
	}

	return(ld);
}

PKI_MEM_STACK *URL_get_data_ldap_url( URL *url, ssize_t size ) {

	LDAP	*ld = NULL;
	int  	i,rc;
	char	*attrs[] = { url->attrs, NULL } ;
	char    *filter = "objectclass=*";
	struct berval **vals = NULL;
	LDAPMessage *res = NULL;
	PKI_MEM_STACK *ret = NULL;
	PKI_MEM *obj = NULL;

	if( (!url) || (!url->addr) || (!url->path)) {
		return NULL;
	}

	if((ld = URL_LDAP_connect ( url )) == NULL ) {
		return NULL;
	};

	/* We search for the exact match, so LDAP_SCOPE_BASE is used here */
	if (( rc = ldap_search_s( ld, url->path, LDAP_SCOPE_BASE,
			filter, attrs, 0, &res )) != LDAP_SUCCESS ) {

		if ( rc != LDAP_NO_SUCH_OBJECT ) {
			PKI_log_err("LDAP: [%s] object not found", url->path);
		}

		goto end;
	}

	i = ldap_count_entries( ld, res );

	if((ret = PKI_STACK_MEM_new()) == NULL ) {
		/* ERROR: Allocating memory */
		goto end;
	}
	if((vals = ldap_get_values_len (ld, res, attrs[0])) != NULL ) {

		for( i=0; vals[i] != NULL; i++ ) {
			if((obj = PKI_MEM_new_null()) == NULL ) {
				goto end;
			}
			if(PKI_MEM_add( obj, (char *) vals[i]->bv_val, 
					vals[i]->bv_len) == PKI_ERR) {
				/* ERROR in memory growth */;
				break;
			}
			PKI_STACK_MEM_push( ret, obj );
		}

		ldap_value_free_len( vals );
		vals = NULL;
	} else {
		return NULL;
	}

end:
	if(res) ldap_msgfree( res );
	if(ld) ldap_unbind( ld );

	return (ret);
}

#endif /* LDAP */
